from typing import Iterable, List

from spacy.language import Language, component
from spacy.tokens import Doc, Span, Token

__all__ = [
    "bunsetu_available",
    "bunsetu_span",
    "bunsetu_spans",
    "bunsetu_phrase_span",
    "bunsetu_phrase_spans",
    "bunsetu_head_list",
    "bunsetu_head_tokens",
    "bunsetu_bi_labels",
    "bunsetu_position_types",
    "BunsetuRecognizer",
    "append_bunsetu_head_dep_suffix",
    "BUNSETU_HEAD_SUFFIX",
    "PHRASE_RELATIONS",
    "POS_PHRASE_MAP",
]


BUNSETU_HEAD_SUFFIX = "_bunsetu"

PHRASE_RELATIONS = ("compound", "nummod", "nmod")

POS_PHRASE_MAP = {
    "NOUN": "NP",
    "NUM": "NP",
    "PRON": "NP",
    "PROPN": "NP",

    "VERB": "VP",

    "ADJ": "ADJP",

    "ADV": "ADVP",

    "CCONJ": "CCONJP",
}


def bunsetu_available(span: Span):
    return "bunsetu_heads" in span.doc.user_data


def bunsetu_head_list(span: Span) -> Iterable[int]:
    doc = span.doc
    heads = doc.user_data["bunsetu_heads"]
    if isinstance(span, Doc):
        return heads
    else:
        start = span.start
        end = span.end
        return [i - start for i in heads if start <= i < end]


def bunsetu_head_tokens(span: Span) -> Iterable[Token]:
    doc = span.doc
    heads = doc.user_data["bunsetu_heads"]
    if isinstance(span, Doc):
        start = 0
        end = len(span)
    else:
        start = span.start
        end = span.end
    return [span[i - start] for i in heads if start <= i < end]


def bunsetu_spans(span: Span) -> Iterable[Span]:
    return [
        bunsetu_span(head) for head in bunsetu_head_tokens(span)
    ]


def bunsetu_span(token: Token) -> Span:
    bunsetu_bi_list = bunsetu_bi_labels(token.doc)
    start = token.i
    end = start + 1
    for idx in range(start, 0, -1):
        if bunsetu_bi_list[idx] == "B":
            start = idx
            break
    else:
        start = 0
    doc_len = len(token.doc)
    for idx in range(end, doc_len):
        if bunsetu_bi_list[idx] == "B":
            end = idx
            break
    else:
        end = doc_len

    doc = token.doc
    return Span(doc, start=start, end=end, label=POS_PHRASE_MAP.get(doc[start:end].root.pos_, ""))


def bunsetu_phrase_spans(span: Span, phrase_relations: Iterable[str] = PHRASE_RELATIONS) -> Iterable[Span]:
    return [
        bunsetu_phrase_span(head, phrase_relations) for head in bunsetu_head_tokens(span)
    ]


def bunsetu_phrase_span(token: Token, phrase_relations: Iterable[str] = PHRASE_RELATIONS) -> Span:
    def _traverse(head, _bunsetu, result):
        for t in head.children:
            if _bunsetu.start <= t.i < _bunsetu.end:
                if t.dep_ in phrase_relations:
                    _traverse(t, _bunsetu, result)
        result.append(head.i)
    bunsetu = bunsetu_span(token)
    phrase_tokens = []
    _traverse(bunsetu.root, bunsetu, phrase_tokens)
    start = min(phrase_tokens)
    end = max(phrase_tokens) + 1
    return Span(token.doc, start=start, end=end, label=bunsetu.label_)


def bunsetu_bi_labels(span: Span) -> List[str]:
    doc = span.doc
    bunsetu_bi = doc.user_data["bunsetu_bi_labels"]
    if isinstance(span, Doc):
        return bunsetu_bi
    else:
        start = span.start
        end = span.end
        return bunsetu_bi[start:end]


def bunsetu_position_types(span: Span) -> List[str]:
    doc = span.doc
    position_types = doc.user_data["bunsetu_position_types"]
    if isinstance(span, Doc):
        return position_types
    else:
        start = span.start
        end = span.end
        return position_types[start:end]


@component("BunsetuRecognizer", requires=("token.dep",), assigns=("token.dep",))
class BunsetuRecognizer:
    def __init__(self, nlp: Language, **_cfg) -> None:
        self.nlp = nlp

    def __call__(self, doc: Doc) -> Doc:
        debug = False
        heads = [False] * len(doc)
        for t in doc:
            if t.dep_ == "ROOT":
                heads[t.i] = True
            elif t.dep_.endswith(BUNSETU_HEAD_SUFFIX):
                heads[t.i] = True
                t.dep_ = t.dep_[:-len(BUNSETU_HEAD_SUFFIX)]
        for t in doc:  # recovering uncovered subtrees
            if heads[t.i]:
                while t.head.i < t.i and not heads[t.head.i]:
                    heads[t.head.i] = t.head.pos_ not in {"PUNCT"}
                    if debug and heads[t.head.i]:
                        print("========= A", t.i + 1, t.orth_, "=========")
                        print(list((t.i + 1, t.orth_, t.head.i + 1) for t, is_head in zip(doc, heads) if is_head))
                    t = t.head
                heads[t.head.i] = True

        for ent in doc.ents:  # removing head inside ents
            head = None
            outer = None
            for t in ent:
                if t.head.i == t.i or t.head.i < ent.start or ent.end <= t.head.i:
                    if not outer:
                        head = t
                        outer = t.head
                    elif outer.i != t.head.i:
                        break
            else:
                if head:
                    for t in ent:
                        if t.i != head.i:
                            heads[t.i] = False

        bunsetu_heads = tuple(idx for idx, is_head in enumerate(heads) if is_head)

        bunsetu_bi = ["I"] * len(doc)
        bunsetu_bi[0] = "B"
        for head_i, next_head_i in zip(bunsetu_heads[:-1], bunsetu_heads[1:]):
            l_head = doc[head_i]
            r_head = doc[next_head_i]
            if l_head.right_edge.i + 1 == r_head.left_edge.i or l_head.right_edge.i >= r_head.i:  # (l)(r) or (l (r))
                bunsetu_bi[r_head.left_edge.i] = "B"
            elif l_head.i <= r_head.left_edge.i:  # ((l) r)
                bunsetu_bi[l_head.right_edge.i + 1] = "B"
            else:  # ((l) (missed_tokens) (r))
                l_ancestors = set(t.i for t in l_head.ancestors)
                r_ancestors = set(t.i for t in r_head.ancestors)
                for m in doc[l_head.right_edge.i + 1: r_head.left_edge.i]:  # find closer branch
                    found = False
                    for m_ancestor in [m] + list(m.ancestors):
                        if m_ancestor.i in r_ancestors:
                            bunsetu_bi[m_ancestor.i] = "B"
                            found = True
                            break
                        elif m_ancestor.i in l_ancestors:
                            break
                    if found:
                        break
                else:
                    bunsetu_bi[l_head.right_edge.i + 1] = "B"

        doc.user_data["bunsetu_heads"] = bunsetu_heads
        doc.user_data["bunsetu_bi_labels"] = bunsetu_bi

        position_types = [None] * len(doc)
        for head in bunsetu_heads:
            phrase = bunsetu_phrase_span(doc[head])
            for t in phrase:
                if t.i == t.head.i:
                    position_types[t.i] = "ROOT"
                elif t.i == head:
                    position_types[t.i] = "NO_HEAD" if t.dep_ == "punct" else "SEM_HEAD"
                else:
                    position_types[t.i] = "CONT"
        first_func = True
        for t, bi, position_type in reversed(list(zip(doc, bunsetu_bi, position_types))):
            if bi:
                first_func = True
            if position_type is None:
                if t.pos_ in {'AUX', 'ADP', 'SCONJ', 'CCONJ', 'PART'}:
                    if first_func:
                        position_types[t.i] = "SYN_HEAD"
                        first_func = False
                    else:
                        position_types[t.i] = "FUNC"
                else:
                    position_types[t.i] = "CONT"
        doc.user_data["bunsetu_position_types"] = position_types
        return doc


def append_bunsetu_head_dep_suffix(tokens: List[Token], suffix: str = BUNSETU_HEAD_SUFFIX) -> None:
    if not suffix:
        return
    for token in tokens:
        if token.dep_.lower() == "root":
            return
        if token.head.i < tokens[0].i or tokens[-1].i < token.head.i:
            token.dep_ += suffix
            return



import plac

from .command_line import run_ginza, run_ginzame


def main_ginzame():
    plac.call(run_ginzame)


def main_ginza():
    plac.call(run_ginza)


if __name__ == "__main__":
    plac.call(run_ginza)



# encoding: utf8
from collections import OrderedDict

import srsly

from spacy import util
from spacy.language import component
from spacy.lang.ja import resolve_pos

__all__ = [
    "CompoundSplitter",
    "tag_to_pos",
]


TAG_DEP_MAP = {
    "ADJ": "amod",
    "ADP": "case",
    "NUM": "nummod",
    "PART": "mark",
    "PUNCT": "punct",
}


def tag_dep_map(tag):
    return TAG_DEP_MAP.get(tag, "compound")


def tag_to_pos(sub_tokens, next_token_tag):
    pos_list = []
    next_pos = None
    for t1, t2 in zip(sub_tokens[:-1], sub_tokens[1:]):
        if next_pos:
            pos = next_pos
            next_pos = None
        else:
            pos, next_pos = resolve_pos(t1.surface, t1.tag, t2.tag)
        pos_list.append(pos)
    if next_pos:
        pos = next_pos
    else:
        pos, next_pos = resolve_pos(sub_tokens[-1].surface, sub_tokens[-1].tag, next_token_tag)
    pos_list.append(pos)
    return pos_list


def _replace_list_entries(lst, index, inserting_list):
    return lst[:index] + inserting_list + lst[index + 1:]


@component("CompoundSplitter", retokenizes=True)
class CompoundSplitter:
    def __init__(self, nlp, **cfg):
        self.nlp = nlp
        self.split_mode = cfg.get("split_mode", None)

    def __call__(self, doc):
        if self._split_mode is None:
            return doc
        elif self._split_mode == "C":
            del doc.user_data["sub_tokens"]
            return doc
        elif self._split_mode == "B":
            sub_tokens_index = 1
        elif self._split_mode == "A":
            sub_tokens_index = 0
        else:
            raise Exception("invalid split_mode: " + self._split_mode)

        sub_tokens_list = [
            sub_tokens[sub_tokens_index] if sub_tokens else None for sub_tokens in doc.user_data["sub_tokens"]
        ]

        for token_i, sub_tokens in reversed(tuple(zip(range(len(doc)), sub_tokens_list))):
            token = doc[token_i]
            token_ent_type = token.ent_type

            # edit token.dep_
            if token.head.i == token.i:
                dep = "ROOT"
            else:
                dep = token.dep_

            compounds = dep in {"compound", "nummod", "punct"}

            # retokenize
            if sub_tokens_index is not None and sub_tokens:
                deps = [tag_dep_map(dtoken.tag) for dtoken in sub_tokens[:-1]] + [token.dep_]
                last = len(sub_tokens) - 1
                if token.head.i == token.i:
                    heads = [(token, last) for _ in range(last + 1)]
                elif compounds:
                    heads = [token.head for _ in range(len(sub_tokens))]
                else:
                    heads = [(token, last) for _ in range(last)] + [token.head]
                surfaces = [dtoken.surface for dtoken in sub_tokens]
                attrs = {
                    "TAG": [dtoken.tag for dtoken in sub_tokens],
                    "DEP": deps,
                    "POS": tag_to_pos(
                        sub_tokens,
                        doc[token.i + 1].tag_ if token.i < len(doc) - 1 else None
                    )
                }
                try:
                    with doc.retokenize() as retokenizer:
                        retokenizer.split(token, surfaces, heads=heads, attrs=attrs)
                except Exception as e:
                    import sys
                    print("Retokenization error:", file=sys.stderr)
                    print(doc.text, file=sys.stderr)
                    print([(t.i, t.orth_) for t in doc], file=sys.stderr)
                    print(list(enumerate(doc.user_data["sub_tokens"])), file=sys.stderr)
                    raise e

                # work-around: retokenize() does not consider the head of the splitted tokens
                if not compounds:
                    for t in doc:
                        if t.i < token_i or token_i + len(sub_tokens) <= t.i:
                            if t.head.i == token_i:
                                t.head = doc[token_i + last]

                for t, dtoken in zip(
                    doc[token_i:token_i + len(sub_tokens)],
                    sub_tokens
                ):
                    t.lemma_ = dtoken.lemma

                if token_ent_type:  # work-around: _retokenize() sets value for Token.ent_iob but not for ent_type
                    for t in doc[token_i + 1:token_i + len(sub_tokens)]:
                        t.ent_type = token_ent_type

                if "inflections" in doc.user_data:
                    doc.user_data["inflections"] = _replace_list_entries(
                        doc.user_data["inflections"],
                        token_i,
                        tuple(dtoken.inf for dtoken in sub_tokens),
                    )
                if "reading_forms" in doc.user_data:
                    doc.user_data["reading_forms"] = _replace_list_entries(
                        doc.user_data["reading_forms"],
                        token_i,
                        tuple(dtoken.reading for dtoken in sub_tokens),
                    )

        del doc.user_data["sub_tokens"]
        return doc

    @property
    def split_mode(self) -> str:
        return self._split_mode

    @split_mode.setter
    def split_mode(self, mode: str):
        assert mode in (None, "A", "B", "C"), 'split_mode should be "A", "B", "C", or None'
        self._split_mode = mode

    def _get_config(self):
        config = OrderedDict(
            (
                ("split_mode", self._split_mode),
            )
        )
        return config

    def _set_config(self, config=None):
        self.split_mode = config.get("split_mode", None) if config else None

    def to_bytes(self, **_kwargs):
        serializers = OrderedDict(
            (
                ("cfg", lambda: srsly.json_dumps(self._get_config())),
            )
        )
        return util.to_bytes(serializers, [])

    def from_bytes(self, data, **_kwargs):
        deserializers = OrderedDict(
            (
                ("cfg", lambda b: self._set_config(srsly.json_loads(b))),
            )
        )
        util.from_bytes(data, deserializers, [])
        return self

    def to_disk(self, path, **_kwargs):
        path = util.ensure_path(path)
        serializers = OrderedDict(
            (
                ("cfg", lambda p: srsly.write_json(p, self._get_config())),
            )
        )
        return util.to_disk(path, serializers, [])

    def from_disk(self, path, **_kwargs):
        path = util.ensure_path(path)
        serializers = OrderedDict(
            (
                ("cfg", lambda p: self._set_config(srsly.read_json(p))),
            )
        )
        util.from_disk(path, serializers, [])



# encoding: utf8

__all__ = [
    "ENE_ONTONOTES_MAPPING",
]


ENE_ONTONOTES_MAPPING = {
    "Person": "PERSON",
    "God": "PERSON",

    "International_Organization": "NORP",
    "Ethnic_Group": "NORP",
    "Ethnic_Group_Other": "NORP",
    "Nationality": "NORP",
    "Political_Organization": "NORP",
    "Political_Organization_Other": "NORP",
    "Political_Party": "NORP",
    "Religion": "NORP",

    "Language": "LANGUAGE",
    "Language_Other": "LANGUAGE",
    "National_Language": "LANGUAGE",

    "Location_Other": "GPE",
    "GPE": "GPE",
    "GPE_Other": "GPE",
    "City": "GPE",
    "Province": "GPE",
    "Country": "GPE",
    "Spa": "GPE",
    "Address": "GPE",
    "Address_Other": "GPE",
    "Postal_Address": "GPE",
    "County": "GPE",

    "Region": "LOC",
    "Region_Other": "LOC",
    "Continental_Region": "LOC",
    "Domestic_Region": "LOC",
    "Geological_Region": "LOC",
    "Geological_Region_Other": "LOC",
    "Mountain": "LOC",
    "Island": "LOC",
    "River": "LOC",
    "Lake": "LOC",
    "Sea": "LOC",
    "Bay": "LOC",

    "Event_Other": "EVENT",
    "Occasion": "EVENT",
    "Occasion_Other": "EVENT",
    "Election": "EVENT",
    "Religious_Festival": "EVENT",
    "Competition": "EVENT",
    "Game": "EVENT",  # used in v7 ("Competition" in v8)
    "Conference": "EVENT",
    "Incident": "EVENT",
    "Incident_Other": "EVENT",
    "War": "EVENT",
    "Natural_Phenomenon": "EVENT",
    "Natural_Phenomenon_Other": "EVENT",
    "Natural_Disaster": "EVENT",
    "Earthquake": "EVENT",

    "Facility": "FAC",
    "Facility_Other": "FAC",
    "Facility_Part": "FAC",
    "Dam": "FAC",
    "Archaeological_Place": "FAC",
    "Archaeological_Place_Other": "FAC",
    "Tomb": "FAC",
    "FOE": "FAC",
    "FOE_Other": "FAC",
    "GOE_Other": "FAC",  # used in v7 ("FOE_Other" in v8)
    "Military_Base": "FAC",
    "Power_Plant": "FAC",
    "Park": "FAC",
    "Shopping_Complex": "FAC",
    "Sports_Facility": "FAC",
    "Museum": "FAC",
    "Zoo": "FAC",
    "Amusement_Park": "FAC",
    "Theater": "FAC",
    "Worship_Place": "FAC",
    "Castle": "FAC",
    "Palace": "FAC",
    "Public_Institution": "FAC",
    "Accommodation": "FAC",
    "Medical_Institution": "FAC",
    "School": "FAC",
    "Research_Institute": "FAC",
    "Market": "FAC",
    "Transport_Facility": "FAC",
    "Transport_Facility_Other": "FAC",
    "Car_Stop": "FAC",
    "Station": "FAC",
    "Airport": "FAC",
    "Port": "FAC",
    "Line": "FAC",
    "Line_Other": "FAC",
    "Railroad": "FAC",
    "Road": "FAC",
    "Canal": "FAC",
    "Water_Route": "FAC",
    "Tunnel": "FAC",
    "Bridge": "FAC",
    "Tumulus": "FAC",

    "Organization": "ORG",
    "Organization_Other": "ORG",
    "Show_Organization": "ORG",
    "Family": "ORG",
    "Sports_Organization": "ORG",
    "Sports_Organization_Other": "ORG",
    "Pro_Sports_Organization": "ORG",  # used in v7 ("Sports_Organization" in v8)
    "Sports_Federation": "ORG",
    "Sports_League": "ORG",
    "Sports_Team": "ORG",
    "Juridical_Person": "ORG",
    "Juridical_Person_Other": "ORG",
    "Channel": "ORG",
    "Corporation_Other": "ORG",  # used in v7 ("Juridical_Person_Other" in v8)
    "Nonprofit_Organization": "ORG",
    "Company": "ORG",
    "Company_Group": "ORG",
    "Government": "ORG",
    "Cabinet": "ORG",
    "Military": "ORG",

    "Product_Other": "PRODUCT",
    "Service": "PRODUCT",
    "Character": "PRODUCT",
    "ID_Number": "PRODUCT",
    "Game_Other": "PRODUCT",
    "Degital_Game": "PRODUCT",
    "Software": "PRODUCT",
    "Vehicle": "PRODUCT",
    "Vehicle_Other": "PRODUCT",
    "Car": "PRODUCT",
    "Train": "PRODUCT",
    "Aircraft": "PRODUCT",
    "Spaceship": "PRODUCT",
    "Ship": "PRODUCT",
    "Food_Other": "PRODUCT",
    "Musical_Instrument": "PRODUCT",
    "Clothing": "PRODUCT",
    "Money_Form": "PRODUCT",
    "Drug": "PRODUCT",
    "Weapon": "PRODUCT",
    "Stock": "PRODUCT",
    "Award": "PRODUCT",
    "Decoration": "PRODUCT",

    "Video_Work": "WORK_OF_ART",
    "Art": "WORK_OF_ART",
    "Art_Other": "WORK_OF_ART",
    "Painting": "WORK_OF_ART",
    "Broadcast_Program": "WORK_OF_ART",
    "Movie": "WORK_OF_ART",
    "Show": "WORK_OF_ART",
    "Music": "WORK_OF_ART",
    "Book": "WORK_OF_ART",
    "Printing": "WORK_OF_ART",
    "Printing_Other": "WORK_OF_ART",
    "Newspaper": "WORK_OF_ART",
    "Magazine": "WORK_OF_ART",
    "Picture": "WORK_OF_ART",

    "Offense": "LAW",
    "Doctrine_Method_Other": "LAW",
    "Movement": "LAW",
    "Plan": "LAW",
    "Rule": "LAW",
    "Rule_Other": "LAW",
    "Treaty": "LAW",
    "Law": "LAW",

    "Timex": "DATE",
    "Timex_Other": "DATE",
    "Timeex": "DATE",
    "Timeex_Other": "DATE",
    "Date": "DATE",
    "Day_Of_Week": "DATE",
    "Era": "DATE",
    "Periodx": "DATE",
    "Periodx_Other": "DATE",
    "Period_Day": "DATE",
    "Period_Week": "DATE",
    "Period_Month": "DATE",
    "Period_Year": "DATE",
    "Time_Top_Other": "DATE",

    "Time": "TIME",
    "Period_Time": "TIME",

    "Percent": "PERCENT",

    "Currency": "MONEY",
    "Money": "MONEY",

    "Unit_Other": "QUANTITY",
    "Latitude_Longitude": "QUANTITY",
    "Latitude_Longtitude": "QUANTITY",  # used in v7 ("Latitude_Longitude" in v8)
    "Measurement": "QUANTITY",
    "Measurement_Other": "QUANTITY",
    "Physical_Extent": "QUANTITY",
    "Seismic_Magnitude": "QUANTITY",
    "Space": "QUANTITY",
    "Volume": "QUANTITY",
    "Weight": "QUANTITY",
    "Speed": "QUANTITY",
    "Intensity": "QUANTITY",
    "Temperature": "QUANTITY",
    "Calorie": "QUANTITY",
    "Seismic_Intensity": "QUANTITY",
    "Countx": "QUANTITY",
    "Countx_Other": "QUANTITY",
    "N_Person": "QUANTITY",
    "N_Organization": "QUANTITY",
    "N_Location": "QUANTITY",
    "N_Location_Other": "QUANTITY",
    "N_Country": "QUANTITY",
    "N_Facility": "QUANTITY",
    "N_Product": "QUANTITY",
    "N_Event": "QUANTITY",
    "N_Natural_Object": "QUANTITY",
    "N_Natural_Object_Other": "QUANTITY",
    "N_Animal": "QUANTITY",
    "N_Flora": "QUANTITY",
    "Point": "QUANTITY",
    "Multiplication": "QUANTITY",
    "Frequency": "QUANTITY",
    "Age": "QUANTITY",

    "Rank": "ORDINAL",
    "School_Age": "ORDINAL",
    "Ordinal_Number": "ORDINAL",

    "Stock_Index": "CARDINAL",

    "Phone_Number": "PHONE",

    "Email": "EMAIL",

    "URL": "URL",

    "Individual_Animal": "ANIMAL",
    "Individual_Animal_Other": "ANIMAL",
    "Racehorse": "ANIMAL",

    "Name": "OTHERS",
    "Name_Other": "OTHERS",
    "Natural_Object": "OTHERS",
    "Natural_Object_Other": "OTHERS",
    "Element": "OTHERS",
    "Compound": "OTHERS",
    "Mineral": "OTHERS",
    "Living_Thing": "OTHERS",
    "Living_Thing_Other": "OTHERS",
    "Fungus": "OTHERS",
    "Mollusk_Arthropod": "OTHERS",
    "Mollusc_Arthropod": "OTHERS",  # used in v7 ("Mollusk_Arthropod" in v8)
    "Insect": "OTHERS",
    "Fish": "OTHERS",
    "Amphibia": "OTHERS",
    "Reptile": "OTHERS",
    "Bird": "OTHERS",
    "Mammal": "OTHERS",
    "Flora": "OTHERS",
    "Living_Thing_Part": "OTHERS",
    "Living_Thing_Part_Other": "OTHERS",
    "Animal_Part": "OTHERS",
    "Flora_Part": "OTHERS",
    "Disease": "OTHERS",
    "Disease_Other": "OTHERS",
    "Animal_Disease": "OTHERS",
    "Color": "OTHERS",
    "Color_Other": "OTHERS",
    "Nature_Color": "OTHERS",
    "Location": "OTHERS",
    "Astronomical_Object": "OTHERS",
    "Astronomical_Object_Other": "OTHERS",
    "Star": "OTHERS",
    "Planet": "OTHERS",
    "Constellation": "OTHERS",
    "Product": "OTHERS",
    "Class": "OTHERS",
    "Food": "OTHERS",
    "Dish": "OTHERS",
    "Doctrine_Method": "OTHERS",
    "Culture": "OTHERS",
    "Academic": "OTHERS",
    "Sport": "OTHERS",
    "Style": "OTHERS",
    "Theory": "OTHERS",
    "Title": "OTHERS",
    "Title_Other": "OTHERS",
    "Position_Vocation": "OTHERS",
    "Unit": "OTHERS",
    "Virtual_Address": "OTHERS",
    "Virtual_Address_Other": "OTHERS",
    "Event": "OTHERS",
    "Numex": "OTHERS",
    "Numex_Other": "OTHERS",
    "Astral_Body_Other": "OTHERS",
    "Material": "OTHERS",
}

ENE8_LABELS = {
    "Name": "1",
    "Name_Other": "1.0",
    "Person": "1.1",
    "God": "1.2",
    "Individual_Animal": "1.3",
    "Individual_Animal_Other": "1.3.0",
    "Racehorse": "1.3.1",
    "Organization": "1.4",
    "Organization_Other": "1.4.0",
    "International_Organization": "1.4.1",
    "Show_Organization": "1.4.2",
    "Family": "1.4.3",
    "Ethnic_Group": "1.4.4",
    "Ethnic_Group_Other": "1.4.4.0",
    "Nationality": "1.4.4.1",
    "Sports_Organization": "1.4.5",
    "Sports_Organization_Other": "1.4.5.0",
    "Sports_Federation": "1.4.5.1",
    "Sports_League": "1.4.5.2",
    "Sports_Team": "1.4.5.3",
    "Juridical_Person": "1.4.6",
    "Juridical_Person_Other": "1.4.6.0",
    "Nonprofit_Organization": "1.4.6.1",
    "Company": "1.4.6.2",
    "Company_Group": "1.4.6.3",
    "Political_Organization": "1.4.7",
    "Political_Organization_Other": "1.4.7.0",
    "Government": "1.4.7.1",
    "Political_Party": "1.4.7.2",
    "Cabinet": "1.4.7.3",
    "Military": "1.4.7.4",
    "Location": "1.5",
    "Location_Other": "1.5.0",
    "GPE": "1.5.1",
    "GPE_Other": "1.5.1.0",
    "City": "1.5.1.1",
    "Province": "1.5.1.2",
    "Country": "1.5.1.3",
    "Region": "1.5.2",
    "Region_Other": "1.5.2.0",
    "Continental_Region": "1.5.2.1",
    "Domestic_Region": "1.5.2.2",
    "Geological_Region": "1.5.3",
    "Geological_Region_Other": "1.5.3.0",
    "Spa": "1.5.3.1",
    "Mountain": "1.5.3.2",
    "Island": "1.5.3.3",
    "River": "1.5.3.4",
    "Lake": "1.5.3.5",
    "Sea": "1.5.3.6",
    "Bay": "1.5.3.7",
    "Astronomical_Object": "1.5.4",
    "Astronomical_Object_Other": "1.5.4.0",
    "Star": "1.5.4.1",
    "Planet": "1.5.4.2",
    "Constellation": "1.5.4.3",
    "Address": "1.5.5",
    "Address_Other": "1.5.5.0",
    "Postal_Address": "1.5.5.1",
    "Facility": "1.6",
    "Facility_Other": "1.6.0",
    "Facility_Part": "1.6.1",
    "Dam": "1.6.2",
    "Archaeological_Place": "1.6.3",
    "Archaeological_Place_Other": "1.6.3.0",
    "Tomb": "1.6.3.1",
    "FOE": "1.6.4",
    "FOE_Other": "1.6.4.0",
    "Military_Base": "1.6.4.1",
    "Castle": "1.6.4.2",
    "Palace": "1.6.4.3",
    "Public_Institution": "1.6.4.4",
    "Accommodation": "1.6.4.5",
    "Medical_Institution": "1.6.4.6",
    "School": "1.6.4.7",
    "Research_Institute": "1.6.4.8",
    "Market": "1.6.4.9",
    "Power_Plant": "1.6.4.10",
    "Park": "1.6.4.11",
    "Shopping_Complex": "1.6.4.12",
    "Sports_Facility": "1.6.4.13",
    "Museum": "1.6.4.14",
    "Zoo": "1.6.4.15",
    "Amusement_Park": "1.6.4.16",
    "Theater": "1.6.4.17",
    "Worship_Place": "1.6.4.18",
    "Transport_Facility": "1.6.5",
    "Transport_Facility_Other": "1.6.5.0",
    "Car_Stop": "1.6.5.1",
    "Station": "1.6.5.2",
    "Airport": "1.6.5.3",
    "Port": "1.6.5.4",
    "Line": "1.6.6",
    "Line_Other": "1.6.6.0",
    "Railroad": "1.6.6.1",
    "Road": "1.6.6.2",
    "Canal": "1.6.6.3",
    "Water_Route": "1.6.6.4",
    "Tunnel": "1.6.6.5",
    "Bridge": "1.6.6.6",
    "Product": "1.7",
    "Product_Other": "1.7.0",
    "Video_Work": "1.7.1",
    "Musical_Instrument": "1.7.2",
    "Clothing": "1.7.3",
    "Money_Form": "1.7.4",
    "Drug": "1.7.5",
    "Weapon": "1.7.6",
    "Stock": "1.7.7",
    "Award": "1.7.8",
    "Decoration": "1.7.9",
    "Offense": "1.7.10",
    "Service": "1.7.11",
    "Class": "1.7.12",
    "Character": "1.7.13",
    "ID_Number": "1.7.14",
    "Game": "1.7.15",
    "Game_Other": "1.7.15.0",
    "Degital_Game": "1.7.15.1",
    "Software": "1.7.16",
    "Vehicle": "1.7.17",
    "Vehicle_Other": "1.7.17.0",
    "Car": "1.7.17.1",
    "Train": "1.7.17.2",
    "Aircraft": "1.7.17.3",
    "Spaceship": "1.7.17.4",
    "Ship": "1.7.17.5",
    "Food": "1.7.18",
    "Food_Other": "1.7.18.0",
    "Dish": "1.7.18.1",
    "Art": "1.7.19",
    "Art_Other": "1.7.19.0",
    "Painting": "1.7.19.1",
    "Broadcast_Program": "1.7.19.2",
    "Movie": "1.7.19.3",
    "Show": "1.7.19.4",
    "Music": "1.7.19.5",
    "Book": "1.7.19.6",
    "Printing": "1.7.20",
    "Printing_Other": "1.7.20.0",
    "Newspaper": "1.7.20.1",
    "Magazine": "1.7.20.2",
    "Doctrine_Method": "1.7.21",
    "Doctrine_Method_Other": "1.7.21.0",
    "Culture": "1.7.21.1",
    "Religion": "1.7.21.2",
    "Academic": "1.7.21.3",
    "Sport": "1.7.21.4",
    "Style": "1.7.21.5",
    "Movement": "1.7.21.6",
    "Theory": "1.7.21.7",
    "Plan": "1.7.21.8",
    "Rule": "1.7.22",
    "Rule_Other": "1.7.22.0",
    "Treaty": "1.7.22.1",
    "Law": "1.7.22.2",
    "Title": "1.7.23",
    "Title_Other": "1.7.23.0",
    "Position_Vocation": "1.7.23.1",
    "Language": "1.7.24",
    "Language_Other": "1.7.24.0",
    "National_Language": "1.7.24.1",
    "Unit": "1.7.25",
    "Unit_Other": "1.7.25.0",
    "Currency": "1.7.25.1",
    "Virtual_Address": "1.8",
    "Virtual_Address_Other": "1.8.0",
    "Channel": "1.8.1",
    "Phone_Number": "1.8.2",
    "Email": "1.8.3",
    "URL": "1.8.4",
    "Event": "1.9",
    "Event_Other": "1.9.0",
    "Occasion": "1.9.1",
    "Occasion_Other": "1.9.1.0",
    "Election": "1.9.1.1",
    "Religious_Festival": "1.9.1.2",
    "Competition": "1.9.1.3",
    "Conference": "1.9.1.4",
    "Incident": "1.9.2",
    "Incident_Other": "1.9.2.0",
    "War": "1.9.2.1",
    "Natural_Phenomenon": "1.9.3",
    "Natural_Phenomenon_Other": "1.9.3.0",
    "Natural_Disaster": "1.9.3.1",
    "Earthquake": "1.9.3.2",
    "Natural_Object": "1.10",
    "Natural_Object_Other": "1.10.0",
    "Element": "1.10.1",
    "Compound": "1.10.2",
    "Mineral": "1.10.3",
    "Living_Thing": "1.10.4",
    "Living_Thing_Other": "1.10.4.0",
    "Fungus": "1.10.4.1",
    "Mollusk_Arthropod": "1.10.4.2",
    "Insect": "1.10.4.3",
    "Fish": "1.10.4.4",
    "Amphibia": "1.10.4.5",
    "Reptile": "1.10.4.6",
    "Bird": "1.10.4.7",
    "Mammal": "1.10.4.8",
    "Flora": "1.10.4.9",
    "Living_Thing_Part": "1.10.5",
    "Living_Thing_Part_Other": "1.10.5.0",
    "Animal_Part": "1.10.5.1",
    "Flora_Part": "1.10.5.2",
    "Disease": "1.11",
    "Disease_Other": "1.11.0",
    "Animal_Disease": "1.11.1",
    "Color": "1.12",
    "Color_Other": "1.12.0",
    "Nature_Color": "1.12.1",
    "Timex": "2",
    "Timex_Other": "2.0",
    "Timeex": "2.1",
    "Timeex_Other": "2.1.0",
    "Time": "2.1.1",
    "Date": "2.1.2",
    "Day_Of_Week": "2.1.3",
    "Era": "2.1.4",
    "Periodx": "2.2",
    "Periodx_Other": "2.2.0",
    "Period_Time": "2.2.1",
    "Period_Day": "2.2.2",
    "Period_Week": "2.2.3",
    "Period_Month": "2.2.4",
    "Period_Year": "2.2.5",
    "Numex": "3",
    "Numex_Other": "3.0",
    "Money": "3.1",
    "Stock_Index": "3.2",
    "Point": "3.3",
    "Percent": "3.4",
    "Multiplication": "3.5",
    "Frequency": "3.6",
    "Age": "3.7",
    "School_Age": "3.8",
    "Ordinal_Number": "3.9",
    "Rank": "3.10",
    "Latitude_Longitude": "3.11",
    "Measurement": "3.12",
    "Measurement_Other": "3.12.0",
    "Physical_Extent": "3.12.1",
    "Space": "3.12.2",
    "Volume": "3.12.3",
    "Weight": "3.12.4",
    "Speed": "3.12.5",
    "Intensity": "3.12.6",
    "Temperature": "3.12.7",
    "Calorie": "3.12.8",
    "Seismic_Intensity": "3.12.9",
    "Seismic_Magnitude": "3.12.10",
    "Countx": "3.13",
    "Countx_Other": "3.13.0",
    "N_Person": "3.13.1",
    "N_Organization": "3.13.2",
    "N_Location": "3.13.3",
    "N_Location_Other": "3.13.3.0",
    "N_Country": "3.13.3.1",
    "N_Facility": "3.13.4",
    "N_Product": "3.13.5",
    "N_Event": "3.13.6",
    "N_Natural_Object": "3.13.7",
    "N_Natural_Object_Other": "3.13.7.0",
    "N_Animal": "3.13.7.1",
    "N_Flora": "3.13.7.2",
}

"""
import json
import sys

if __name__ == "__main__":
    for ne, idx in ENE8_LABELS.items():
        if ne not in ENE_NE_MAPPING:
            print(idx, ne, "not in mapping")
    with open(sys.argv[1], "r") as f:
        meta_json = json.load(f)
    for ne in meta_json["labels"]["ner"]:
        if ne not in ENE_NE_MAPPING:
            print(ne, "not in mapping")
    for ent, idx in ENE_NE_MAPPING.items():
        if ent not in ENE8_LABELS and ent not in meta_json["labels"]["ner"]:
            print(idx, ent, "not used")

    for ne, idx, ent in sorted([
        (ne, ENE8_LABELS[ent] if ent in ENE8_LABELS else "_", ent) for ent, ne in ENT_NE_MAPPING.items()
    ]):
        print("\t"{}": "{}",".format(ent, ne))
"""



from functools import singledispatch
from typing import Callable, Iterable, Union, Tuple, TypeVar

from spacy.lang.ja import DetailedToken
from spacy.language import Language
from spacy.tokens import Doc, Span, Token

from .bunsetu_recognizer import *
from .compound_splitter import *
from .ene_ontonotes_mapper import ENE_ONTONOTES_MAPPING


__all__ = [
    "set_split_mode",
    "token_i", "text", "text_with_ws", "orth", "orth_",
    "ent_type", "ent_type_", "ent_iob", "ent_iob_",
    "lemma", "lemma_", "norm", "norm_",
    "pos", "pos_", "tag", "tag_", "dep", "dep_",
    "is_sent_start", "is_stop", "is_not_stop",
    "ent_label_ene", "ent_label_ontonotes",
    "reading_form", "inflection",
    "bunsetu_bi_label", "bunsetu_position_type", "is_bunsetu_head",
    "SEP", "default_join_func",
    "traverse",
    "head", "ancestors", "conjuncts", "children", "lefts", "rights", "subtree",
    "bunsetu", "phrase", "sub_phrases", "phrases",
    "sub_tokens",
    # from bunsetu_recognizer
    "bunsetu_span",
    "bunsetu_spans",
    "bunsetu_phrase_span",
    "bunsetu_phrase_spans",
    "bunsetu_head_list",
    "bunsetu_head_tokens",
    "bunsetu_bi_labels",
    "bunsetu_position_types",
    "BunsetuRecognizer",
    # from compound_splitter
    "CompoundSplitter",
    "tag_to_pos",
]


def set_split_mode(nlp: Language, mode: str):
    splitter = nlp.get_pipe("CompoundSplitter")
    splitter.split_mode = mode


# token field getters

def token_i(token: Token) -> int:
    return token.i


def text(token: Token) -> str:
    return token.text


def text_with_ws(token: Token) -> str:
    return token.text_with_ws


def orth(token: Token) -> int:
    return token.orth


def orth_(token: Token) -> str:
    return token.orth_


def ent_type(token: Token) -> int:
    return token.ent_type


def ent_type_(token: Token) -> str:
    return ENE_ONTONOTES_MAPPING.get(token.ent_type_, "OTHERS")


def ent_iob(token: Token) -> int:
    return token.ent_iob


def ent_iob_(token: Token) -> str:
    return token.ent_iob_


def lemma(token: Token) -> int:
    return token.lemma


def lemma_(token: Token) -> str:
    return token.lemma_


def norm(token: Token) -> int:
    return token.norm


def norm_(token: Token) -> str:
    return token.norm_


def pos(token: Token) -> int:
    return token.pos


def pos_(token: Token) -> str:
    return token.pos_


def tag(token: Token) -> int:
    return token.tag


def tag_(token: Token) -> str:
    return token.tag_


def dep(token: Token) -> int:
    return token.dep


def dep_(token: Token) -> str:
    return token.dep_


def is_sent_start(token: Token) -> bool:
    return token.is_sent_start


def is_stop(token: Token) -> bool:
    return token.is_stop


def is_not_stop(token: Token) -> bool:
    return not token.is_stop


def ent_label_ene(token: Token) -> str:
    if token.ent_iob_ in "BI":
        return token.ent_iob_ + "-" + token.ent_type_
    else:
        return token.ent_iob_


def ent_label_ontonotes(token: Token) -> str:
    if token.ent_iob_ in "BI":
        return token.ent_iob_ + "-" + ENE_ONTONOTES_MAPPING.get(token.ent_type_, "OTHERS")
    else:
        return token.ent_iob_


# token field getters for Doc.user_data

def reading_form(token: Token, use_orth_if_none=True) -> str:
    reading = token.doc.user_data["reading_forms"][token.i]
    if not reading and use_orth_if_none:
        reading = token.orth_
    return reading


def inflection(token: Token) -> str:
    return token.doc.user_data["inflections"][token.i]


# bunsetu related field getters for Doc.user_data

def bunsetu_bi_label(token: Token):
    return bunsetu_bi_labels(token.doc)[token.i]


def bunsetu_position_type(token: Token):
    return bunsetu_position_types(token.doc)[token.i]


def is_bunsetu_head(token: Token):
    return token.i in token.doc.user_data["bunsetu_heads"]


SEP = "+"


def default_join_func(elements):
    return SEP.join([element if isinstance(element, str) else str(element) for element in elements])


T = TypeVar('T')
U = TypeVar('U')
V = TypeVar('V')


# curried function: ex. traverse(children, lemma_)(token)
@singledispatch
def traverse(
        traverse_func: Callable[[Token], Iterable[Token]],
        element_func: Callable[[Token], T] = lambda token: token,
        condition_func: Callable[[Token], bool] = lambda token: True,
        join_func: Callable[[Iterable[T]], U] = lambda lst: lst,
) -> Callable[[Union[Token, Span]], U]:
    return lambda token: join_func([
        element_func(t) for t in traverse_func(token) if condition_func(t)
    ])


# overload: ex. traverse(token, children, lemma_)
@traverse.register(Token)
def _traverse(
        token: Token,
        traverse_func: Callable[[Token], Iterable[Token]],
        element_func: Callable[[Token], T] = lambda token: token,
        condition_func: Callable[[Token], bool] = lambda token: True,
        join_func: Callable[[Iterable[T]], U] = lambda lst: lst,
) -> U:
    return traverse(traverse_func, element_func, condition_func, join_func)(token)


def head(token: Token) -> Token:
    return token.head


def ancestors(token: Token) -> Iterable[Token]:
    return token.ancestors


def conjuncts(token: Token) -> Tuple[Token]:
    return token.conjuncts


def children(token: Token) -> Iterable[Token]:
    return token.children


def lefts(token: Token) -> Iterable[Token]:
    return token.lefts


def rights(token: Token) -> Iterable[Token]:
    return token.rights


def subtree(token: Token) -> Iterable[Token]:
    return token.subtree


# curried function: ex. bunsetu(lemma_)(token)
@singledispatch
def bunsetu(
        element_func: Callable[[Token], T] = lambda token: token,
        condition_func: Callable[[Token], bool] = lambda token: True,
        join_func: Callable[[Iterable[T]], U] = default_join_func,
) -> Callable[[Token], U]:
    return traverse(bunsetu_span, element_func, condition_func, join_func)


# overload: ex. bunsetu(token, lemma_)
@bunsetu.register(Token)
def _bunsetu(
        token: Token,
        element_func: Callable[[Token], T] = lambda token: token,
        condition_func: Callable[[Token], bool] = lambda token: True,
        join_func: Callable[[Iterable[T]], U] = default_join_func,
) -> U:
    return traverse(bunsetu_span, element_func, condition_func, join_func)(token)


# curried function: ex. phrase(lemma_)(token)
@singledispatch
def phrase(
        element_func: Callable[[Token], T] = lambda token: token,
        condition_func: Callable[[Token], bool] = lambda token: True,
        join_func: Callable[[Iterable[T]], U] = default_join_func,
) -> Callable[[Token], U]:
    return traverse(bunsetu_phrase_span, element_func, condition_func, join_func)


# overload: ex. phrase(token)
@phrase.register(Token)
def _phrase(
        token: Token,
        element_func: Callable[[Token], T] = lambda token: token,
        condition_func: Callable[[Token], bool] = lambda token: True,
        join_func: Callable[[Iterable[T]], U] = default_join_func,
) -> U:
    return traverse(bunsetu_phrase_span, element_func, condition_func, join_func)(token)


# curried function: ex. sub_phrases(lemma_)(token)
@singledispatch
def sub_phrases(
        phrase_func: Callable[[Token], U] = _phrase,
        condition_func: Callable[[Token], bool] = lambda token: True,
) -> Callable[[Token], Iterable[Tuple[str, U]]]:
    return lambda token: _sub_phrases(
        token,
        phrase_func,
        condition_func,
    )


# overload: ex. sub_phrases(token, lemma_)
@sub_phrases.register(Token)
def _sub_phrases(
        token: Token,
        phrase_func: Callable[[Token], U] = _phrase,
        condition_func: Callable[[Token], bool] = lambda token: True,
) -> Iterable[Tuple[str, U]]:
    return [
        (
            t.dep_,
            phrase_func(t),
        ) for t in bunsetu_span(token).root.children if t.i in bunsetu_head_list(token.doc) and condition_func(t)
    ]


# curried function: ex. phrases(lemma_)(sent)
@singledispatch
def phrases(
        phrase_func: Callable[[Token], U] = _phrase,
        condition_func: Callable[[Token], bool] = lambda token: True,
) -> Callable[[Span], Iterable[U]]:
    return lambda sent: _phrases_span(
        sent,
        phrase_func,
        condition_func,
    ) if isinstance(sent, Span) else _phrases_doc(
        sent,
        phrase_func,
        condition_func,
    )


# overload: ex. phrases(sent, lemma_)
@phrases.register(Span)
def _phrases_span(
        sent: Span,
        phrase_func: Callable[[Token], U] = _phrase,
        condition_func: Callable[[Token], bool] = lambda token: True,
) -> Iterable[U]:
    return [
        phrase_func(t) for t in bunsetu_head_tokens(sent) if condition_func(t)
    ]


# overload: ex. phrases(doc, lemma_)
@phrases.register(Doc)
def _phrases_doc(
        doc: Doc,
        phrase_func: Callable[[Token], U] = _phrase,
        condition_func: Callable[[Token], bool] = lambda token: True,
) -> Iterable[U]:
    return [
        phrase_func(t) for t in bunsetu_head_tokens(doc[:]) if condition_func(t)
    ]


# curried function: ex. sub_tokens("B", lambda sub_token: sub_token.lemma)(token)
@singledispatch
def sub_tokens(
        mode: str = "A",  # "A" or "B"
        sub_token_func: Callable[[DetailedToken], T] = lambda sub_token: sub_token,
        join_func: Callable[[Iterable[T]], U] = default_join_func,
) -> Callable[[Token], U]:
    return lambda token: _sub_tokens(token, mode, sub_token_func, join_func)


# overload: ex. sub_tokens(token, "B", lambda sub_token: sub_token.lemma)
@sub_tokens.register(Token)
def _sub_tokens(
        token: Token,
        mode: str = "A",  # "A" or "B"
        sub_token_func: Callable[[DetailedToken], T] = lambda sub_token: sub_token.surface,
        join_func: Callable[[Iterable[T]], U] = default_join_func,
) -> U:
    if token.doc.user_data["sub_tokens"][token.i]:
        elements = token.doc.user_data["sub_tokens"][token.i][{"A": 0, "B": 1}[mode]]
    else:
        elements = [
            DetailedToken(
                token.orth_,
                token.tag_,
                inflection(token),
                token.lemma_,
                reading_form(token),
                None,
            )
        ]
    return join_func([
        sub_token_func(element) for element in elements
    ])



# coding: utf8
import json
from multiprocessing import Pool, cpu_count
from pathlib import Path
import sys

import plac
import spacy
from spacy.tokens import Span

from spacy.lang.ja import JapaneseDefaults

from . import set_split_mode, inflection, reading_form, ent_label_ene, ent_label_ontonotes,\
    bunsetu_bi_label, bunsetu_position_type
from .bunsetu_recognizer import bunsetu_available, bunsetu_head_list, bunsetu_phrase_span

MINI_BATCH_SIZE = 100


def run(
        model_path=None,
        split_mode=False,
        hash_comment="print",
        output_path=None,
        output_format="0",
        require_gpu=False,
        parallel=1,
        files=None,
):
    if require_gpu:
        print("GPU enabled", file=sys.stderr)

    analyzer = Analyzer(
        model_path,
        split_mode,
        hash_comment,
        output_format,
        require_gpu,
    )

    if parallel <= 0:
        parallel = max(1, cpu_count() + parallel)

    pool = None

    if output_path:
        output = open(str(output_path), "w")
    else:
        output = sys.stdout

    output_json = output_format in ["3", "json"]
    output_json_opened = False

    def output_json_open():
        nonlocal output_json_opened
        if output_json and not output_json_opened:
            print("[", file=output)
            output_json_opened = True
        elif output_json:
            print(" ,", file=output)

    try:
        if not files:
            if sys.stdin.isatty():
                parallel = 1
            else:
                files = [0]

        if not files:
            analyzer.set_nlp()
            while True:
                line = input()
                for sent in analyzer.analyze_line(line):
                    for ol in sent:
                        output_json_open()
                        print(ol, file=output)
        elif parallel == 1:
            analyzer.set_nlp()
            for path in files:
                with open(path, "r") as f:
                    for line in f:
                        for sent in analyzer.analyze_line(line):
                            for ol in sent:
                                output_json_open()
                                print(ol, file=output)
        else:
            buffer = []
            for file_idx, path in enumerate(files):
                with open(path, "r") as f:
                    while True:
                        eof, buffer = fill_buffer(f, MINI_BATCH_SIZE * parallel, buffer)
                        if eof and (file_idx + 1 < len(files) or len(buffer) == 0):
                            break  # continue to next file
                        if not pool:
                            if len(buffer) <= MINI_BATCH_SIZE:  # enough for single process
                                analyzer.set_nlp()
                                for line in buffer:
                                    for sent in analyzer.analyze_line(line):
                                        for ol in sent:
                                            output_json_open()
                                            print(ol, file=output)
                                break  # continue to next file
                            parallel = (len(buffer) - 1) // MINI_BATCH_SIZE + 1
                            pool = Pool(parallel)

                        mini_batch_size = (len(buffer) - 1) // parallel + 1
                        mini_batches = [
                            buffer[idx * mini_batch_size:(idx + 1) * mini_batch_size] for idx in range(parallel)
                        ]
                        for mini_batch_result in pool.map(analyzer.analyze_lines_mp, mini_batches):
                            for sents in mini_batch_result:
                                for lines in sents:
                                    for ol in lines:
                                        output_json_open()
                                        print(ol, file=output)

                        buffer.clear()  # process remaining part of current file

    except EOFError:
        pass
    except KeyboardInterrupt:
        pass
    finally:
        try:
            if pool:
                pool.close()
        finally:
            if output_json and output_json_opened:
                print("]", file=output)
            output.close()


def fill_buffer(f, batch_size, buffer=None):
    if buffer is None:
        buffer = []

    for line in f:
        buffer.append(line)
        if len(buffer) == batch_size:
            return False, buffer
    return True, buffer


class Analyzer:
    def __init__(
            self,
            model_path,
            split_mode,
            hash_comment,
            output_format,
            require_gpu,
    ):
        self.model_path = model_path
        self.split_mode = split_mode
        self.hash_comment = hash_comment
        self.output_format = output_format
        self.require_gpu = require_gpu
        self.nlp = None

    def set_nlp(self):
        if self.nlp:
            return

        if self.require_gpu:
            spacy.require_gpu()

        if self.output_format in ["2", "mecab"]:
            nlp = JapaneseDefaults.create_tokenizer(config={
                "split_mode": self.split_mode
            }).tokenizer
        else:
            # Work-around for pickle error. Need to share model data.
            if self.model_path:
                nlp = spacy.load(self.model_path)
            else:
                nlp = spacy.load("ja_ginza")

            if self.split_mode:
                set_split_mode(nlp, self.split_mode)

        self.nlp = nlp

    def analyze_lines_mp(self, lines):
        self.set_nlp()
        return tuple(self.analyze_line(line) for line in lines)

    def analyze_line(self, line):
        return analyze(self.nlp, self.hash_comment, self.output_format, line)


def analyze(nlp, hash_comment, output_format, line):
    line = line.rstrip("\n")
    if line.startswith("#"):
        if hash_comment == "print":
            return line,
        elif hash_comment == "skip":
            return (),
    if line == "":
        return "",
    if output_format in ["0", "conllu"]:
        doc = nlp(line)
        return [analyze_conllu(sent) for sent in doc.sents]
    elif output_format in ["1", "cabocha"]:
        doc = nlp(line)
        return [analyze_cabocha(sent) for sent in doc.sents]
    elif output_format in ["2", "mecab"]:
        doc = nlp.tokenize(line)
        return [analyze_mecab(doc)]
    elif output_format in ["3", "json"]:
        doc = nlp(line)
        return [analyze_json(sent) for sent in doc.sents]
    else:
        raise Exception(output_format + " is not supported")


def analyze_json(sent: Span):
    tokens = []
    for token in sent:
        t = {
            "id": token.i - sent.start + 1,
            "orth": token.orth_,
            "tag": token.tag_,
            "pos": token.pos_,
            "lemma": token.lemma_,
            "head": token.head.i - token.i,
            "dep": token.dep_,
            "ner": "{}-{}".format(token.ent_iob_, token.ent_type_) if token.ent_type_ else token.ent_iob_,
        }
        if token.whitespace_:
            t["whitespace"] = token.whitespace_
        tokens.append("       " + json.dumps(t, ensure_ascii=False))
    tokens = ",\n".join(tokens)

    yield """ {{
  "paragraphs": [
   {{
    "raw": "{}",
    "sentences": [
     {{
      "tokens": [
{}
      ]
     }}
    ]
   }}
  ]
 }}""".format(
        sent.text,
        tokens,
    )


def analyze_conllu(sent: Span, print_origin=True):
    if print_origin:
        yield "# text = {}".format(sent.text)
    np_labels = [""] * len(sent)
    use_bunsetu = bunsetu_available(sent)
    if use_bunsetu:
        for head_i in bunsetu_head_list(sent):
            bunsetu_head_token = sent[head_i]
            phrase = bunsetu_phrase_span(bunsetu_head_token)
            if phrase.label_ == "NP":
                for idx in range(phrase.start - sent.start, phrase.end - sent.start):
                    np_labels[idx] = "NP_B" if idx == phrase.start else "NP_I"
    for token, np_label in zip(sent, np_labels):
        yield conllu_token_line(sent, token, np_label, use_bunsetu)
    yield ""


def conllu_token_line(sent, token, np_label, use_bunsetu):
    bunsetu_bi = bunsetu_bi_label(token) if use_bunsetu else None
    position_type = bunsetu_position_type(token) if use_bunsetu else None
    inf = inflection(token)
    reading = reading_form(token)
    ne = ent_label_ontonotes(token)
    ene = ent_label_ene(token)
    misc = "|".join(filter(lambda s: s, (
        "SpaceAfter=Yes" if token.whitespace_ else "SpaceAfter=No",
        "" if not bunsetu_bi else "BunsetuBILabel={}".format(bunsetu_bi),
        "" if not position_type else "BunsetuPositionType={}".format(position_type),
        np_label,
        "" if not inf else "Inf={}".format(inf),
        "" if not reading else "Reading={}".format(reading.replace("|", "\\|").replace("\\", "\\\\")),
        "" if not ne or ne == "O" else "NE={}".format(ne),
        "" if not ene or ene == "O" else "ENE={}".format(ene),
    )))

    return "\t".join([
        str(token.i - sent.start + 1),
        token.orth_,
        token.lemma_,
        token.pos_,
        token.tag_.replace(",*", "").replace(",", "-"),
        "NumType=Card" if token.pos_ == "NUM" else "_",
        "0" if token.head.i == token.i else str(token.head.i - sent.start + 1),
        token.dep_.lower() if token.dep_ else "_",
        "_",
        misc if misc else "_",
    ])


def analyze_cabocha(sent: Span):
    bunsetu_index_list = {}
    bunsetu_index = -1
    for token in sent:
        if bunsetu_bi_label(token) == "B":
            bunsetu_index += 1
        bunsetu_index_list[token.i] = bunsetu_index

    lines = []
    for token in sent:
        if bunsetu_bi_label(token) == "B":
            lines.append(cabocha_bunsetu_line(sent, bunsetu_index_list, token))
        lines.append(cabocha_token_line(token))
    lines.append("EOS")
    lines.append("")
    return lines


def cabocha_bunsetu_line(sent: Span, bunsetu_index_list, token):
    bunsetu_head_index = None
    bunsetu_dep_index = None
    bunsetu_func_index = None
    dep_type = "D"
    for t in token.doc[token.i:sent.end]:
        if bunsetu_index_list[t.i] != bunsetu_index_list[token.i]:
            if bunsetu_func_index is None:
                bunsetu_func_index = t.i - token.i
            break
        tbi = bunsetu_index_list[t.head.i]
        if bunsetu_index_list[t.i] != tbi:
            bunsetu_head_index = t.i - token.i
            bunsetu_dep_index = tbi
        if bunsetu_func_index is None and bunsetu_position_type(t) in {"FUNC", "SYN_HEAD"}:
            bunsetu_func_index = t.i - token.i
    else:
        if bunsetu_func_index is None:
            bunsetu_func_index = len(sent) - token.i
    if bunsetu_head_index is None:
        bunsetu_head_index = 0
    if bunsetu_dep_index is None:
        bunsetu_dep_index = -1
    return "* {} {}{} {}/{} 0.000000".format(
        bunsetu_index_list[token.i],
        bunsetu_dep_index,
        dep_type,
        bunsetu_head_index,
        bunsetu_func_index,
    )


def cabocha_token_line(token):
    part_of_speech = token.tag_.replace("-", ",")
    part_of_speech += ",*" * (3 - part_of_speech.count(",")) + "," + inflection(token)
    reading = reading_form(token)
    return "{}\t{},{},{},{}\t{}".format(
        token.orth_,
        part_of_speech,
        token.lemma_,
        reading if reading else token.orth_,
        "*",
        "O" if token.ent_iob_ == "O" else "{}-{}".format(token.ent_iob_, token.ent_type_),
    )


def analyze_mecab(sudachipy_tokens):
    return tuple(mecab_token_line(t) for t in sudachipy_tokens) + ("EOS", "")


def mecab_token_line(token):
    reading = token.reading_form()
    return "{}\t{},{},{},{}".format(
        token.surface(),
        ",".join(token.part_of_speech()),
        token.normalized_form(),
        reading if reading else token.surface(),
        "*",
    )


@plac.annotations(
    model_path=("model directory path", "option", "b", str),
    split_mode=("split mode", "option", "s", str, ["A", "B", "C", None]),
    hash_comment=("hash comment", "option", "c", str, ["print", "skip", "analyze"]),
    output_path=("output path", "option", "o", Path),
    parallel=("parallel level (default=-1, all_cpus=0)", "option", "p", int),
    files=("input files", "positional"),
)
def run_ginzame(
        model_path=None,
        split_mode=None,
        hash_comment="print",
        output_path=None,
        parallel=-1,
        *files,
):
    run(
        model_path=model_path,
        split_mode=split_mode,
        hash_comment=hash_comment,
        output_path=output_path,
        output_format="mecab",
        require_gpu=False,
        parallel=parallel,
        files=files,
    )


def main_ginzame():
    plac.call(run_ginzame)


@plac.annotations(
    model_path=("model directory path", "option", "b", str),
    split_mode=("split mode", "option", "s", str, ["A", "B", "C", None]),
    hash_comment=("hash comment", "option", "c", str, ["print", "skip", "analyze"]),
    output_path=("output path", "option", "o", Path),
    output_format=("output format", "option", "f", str, ["0", "conllu", "1", "cabocha", "2", "mecab", "3", "json"]),
    require_gpu=("enable require_gpu", "flag", "g"),
    parallel=("parallel level (default=1, all_cpus=0)", "option", "p", int),
    files=("input files", "positional"),
)
def run_ginza(
        model_path=None,
        split_mode=None,
        hash_comment="print",
        output_path=None,
        output_format="conllu",
        require_gpu=False,
        parallel=1,
        *files,
):
    run(
        model_path=model_path,
        split_mode=split_mode,
        hash_comment=hash_comment,
        output_path=output_path,
        output_format=output_format,
        require_gpu=require_gpu,
        parallel=parallel,
        files=files,
    )


def main_ginza():
    plac.call(run_ginza)


if __name__ == "__main__":
    plac.call(run_ginza)



from setuptools import setup, find_packages


setup(
    author="Megagon Labs, Tokyo.",
    author_email="ginza@megagon.ai",
    description="GiNZA, An Open Source Japanese NLP Library, based on Universal Dependencies",
    entry_points={
        "spacy_factories": [
            "BunsetuRecognizer = ginza:BunsetuRecognizer",
            "CompoundSplitter = ginza:CompoundSplitter",
        ],
        "console_scripts": [
            "ginza = ginza.command_line:main_ginza",
            "ginzame = ginza.command_line:main_ginzame",
        ],
    },
    install_requires=[
        "spacy>=2.3.2,<3.0.0",
        "SudachiPy>=0.4.9 ; python_version >= '3.5'",
        "SudachiDict-core>=20200330 ; python_version >= '3.5'",
        "ja_ginza>=4.0.0,<4.1.0",
    ],
    license="MIT",
    name="ginza",
    packages=find_packages(include=["ginza"]),
    url="https://github.com/megagonlabs/ginza",
    version='4.0.4',
)
